# Importing necessary modules including FactSet - SDK
import ipysheet
import pandas as pd
import numpy as np
import math
import os
import ast
import textwrap
import ipywidgets as widgets
from IPython.display import display
import matplotlib.pyplot as plt
from PIL import Image
import fds.sdk.OpenRisk
from fds.sdk.OpenRisk.api import operations_api
from fds.sdk.OpenRisk.models import *


# Establishing a class to define functions for Manager Rebalancing
class ManagerRebalancing:
    def __init__(
        self, username, apikey, manager_structures, manager_mapping, currency, decimal
    ):
        """
        :param username: Username for client's authentication - FDS Username
        :param apikey: apikey associated with FDS Username
        :param manager_structures: CSV file containing manager structures ['PortfolioName','portfolio','benchmark']
        :param manager_mapping: CSV file containing csv file path for respective portfolios ['portfolio','filepath','filename']
        :param currency: string mentioning the base currency
        :param decimal: integer mentioning number of decimal places to display in the report
        """
        self.username = username
        self.apikey = apikey
        self.manager_structures = manager_structures
        self.manager_mapping = manager_mapping
        self.risk_free_cash = (
            "CASH_" + currency
        )  # Adding CASH_ prefix to risk free cash
        self.dis_decimal = decimal
        self.ids = "ids"  # Initialising to keep the text format same across the file
        self.marketValues = "marketValues"
        self.compositeAssetId = "compositeAssetId"
        self.managers = "Managers"
        self.weights = "Weights"
        self.tradingwgt = "Trading Weights"
        self.openrisk = OpenRiskEngine(
            self.username, self.apikey
        )  # Importing OpenRiskEngine class

    def composite_asset_builder_manager_structure(self):
        """
        Constructs stacked manager structures file containing compositeAssetId, ids and marketValues.
        Returns:
                Dataframe: stacked flat file containing compositeAssetId, ids and marketValues.
        """
        try:
            df_structures = pd.read_csv(
                f"{self.manager_structures}"
            )  # Importing managers structure
            df_structures.set_index("PortfolioName", inplace=True)
            self.manager_names = df_structures.index.values.tolist()
            self.manager_holdings_files = df_structures.to_numpy().flatten().tolist()

            df_structures["risk_free_cash"] = [self.risk_free_cash] * len(df_structures)

            # convert manager structures to flat file - df_structure_stacked
            df_structure_stacked = pd.DataFrame()
            df_structure_stacked[self.ids] = pd.concat(
                [
                    df_structures["portfolio"],
                    df_structures["benchmark"],
                    df_structures["risk_free_cash"],
                ]
            )
            df_structure_stacked[self.marketValues] = np.array(
                [100.0] * len(df_structures["portfolio"])
                + [-100.0] * len(df_structures["benchmark"])
                + [100.0] * len(df_structures["risk_free_cash"])
            )
            df_structure_stacked.reset_index(inplace=True)
            df_structure_stacked.rename(
                columns={"PortfolioName": self.compositeAssetId}, inplace=True
            )

            return df_structure_stacked
        except Exception as e:
            print(
                f"Check the input field names in {self.manager_structures}. It should be PortfolioName, portfolio and benchmark"
            )

    def composite_asset_builder_flat_to_esdk_convert(self, ca_expand_lvl):
        """
        Parses into various csv holdings file to add holding composite to existing managers structure flat file.
        Converts the flat file into ESDK compatible composite asset format.
        Args:
            ca_expand_lvl(int): Composite asset expand level
        Returns:
                sdk.composite_asset_definitions
        """
        df_manager_structures = (
            self.composite_asset_builder_manager_structure()
        )  # Bring the manager structures to add with holdings structures
        ca_flat = self.openrisk.composite_asset_builder_expansion(
            df_manager_structures, self.manager_mapping, None, ca_expand_lvl
        )  # expanding the Composite assets
        composite_asset = self.openrisk.composite_asset_builder_esdk(ca_flat)

        return composite_asset

    def beautify(self, df_manager_risk):
        """
        To display only portfolio level and security level values for any risk stats in a report format
        Args:
            df_manager_risk(dataframe): Readable output report converted from SDK output
        Returns:
                Dataframe: Beautified risk report
        """
        risk_port = df_manager_risk["portfolio"]
        risk_port.rename(columns={"portfolio": "Managers"}, inplace=True)
        risk_asset = df_manager_risk["security"]
        risk_asset.rename(columns={"symbol": "Managers"}, inplace=True)
        multi_manager = pd.concat([risk_port, risk_asset], ignore_index=True)
        multi_manager.set_index("Managers", inplace=True)
        multi_manager = multi_manager.copy(deep=True)

        return multi_manager

    def remapping_beautify(self, output_dict, stat_remapping_file, report_name=None):
        """
        Remap the columns and beautify the report
        Arg:
            Output_dict(dict): ESDK output response to calculations
            stat_remapping_file(str): csv path where file contains information on mapping statistics
            report_name(str): Name of the report (optional)
        Returns:
                Dataframe: Remapped and beautified report
        """
        df_manager_risk = self.openrisk.stat_remap(output_dict, stat_remapping_file)
        multi_manager = self.beautify(df_manager_risk)
        float_cols = [
            x for x in multi_manager.columns if multi_manager[x].dtype == "float64"
        ]
        formats = {x: lambda v: f"{v:,.{self.dis_decimal}f}" for x in float_cols}
        if report_name is not None:
            multi_manager_styled = multi_manager.style.set_caption(
                f"<h2 style='font-weight:bold; font-size:16px;'>{report_name}</h2>"
            ).format(formats)
        else:
            multi_manager_styled = multi_manager.style.set_caption(
                f"<h2 style='font-weight:bold; font-size:16px;'></h2>"
            ).format(formats)

        return multi_manager_styled

    def manager_weight_simulation(self):
        """
        Creates an Ipysheet where user provides weights
        Returns:
                ipysheet: sheet containing portfolio and weights
        """
        try:
            managers = pd.read_csv(self.manager_structures)
            managers = managers["PortfolioName"].values.tolist()
            df = pd.DataFrame(
                {self.managers: managers, self.weights: [0] * len(managers)}
            )
            weights_sheet = ipysheet.from_dataframe(df)
            for cell in weights_sheet.cells:
                if cell.column_start == 0:
                    cell.read_only = True

            return weights_sheet
        except Exception as e:
            print(
                print(
                    f"Check the input field names in {self.manager_structures}. It should be PortfolioName, portfolio and benchmark"
                )
            )
            raise e

    def get_weights(self, manager_weights):
        """
        To fetch the weights from ipysheet
        Arg:
            manager_weights(dataframe): dataframe containing portfolio and weights
        Returns:
                Dict: containing ids and marketValues
        """
        manager_weights.replace("", 0.0, inplace=True)
        port_holdings = {
            "ids": manager_weights["Managers"].values.tolist(),
            "market_values": manager_weights["Weights"].astype(float).values.tolist(),
        }
        return port_holdings

    def get_current_weights(self, df_manager_risk, df_stat):
        """
        Fetches the current weights from the report
        Args:
            df_manager_risk(df): Output report converted from SDK output
            df_stat(df): stats object, should contain stat_name and stat_level
        Returns:
            list: list of current weights
        """
        df_manager_btfied = self.beautify(df_manager_risk)
        if "PortfolioPercentWeights" in df_stat["name"].tolist():
            weights = df_manager_btfied["PortfolioPercentWeights"].values.tolist()
        elif "PortfolioWeights" in df_stat["name"].tolist():
            weights = df_manager_btfied["PortfolioWeights"].values.tolist()

        return weights

    def get_bh_stats(self, df_manager_risk, df_stat):
        """
        Fetches the latest best-hedge stats from the report
        Args:
            df_manager_risk(df): Output report converted from SDK output
            df_stat(df): stats object, should contain stat_name and stat_level
        Returns:
            DataFrame: df with sorted data based on Simulated TE
            float: current portfolio TE
        """
        df_manager_btfied = self.beautify(df_manager_risk)
        cur_port_te = df_manager_btfied.loc["portfolio", "ActiveVarianceToRiskCSR"]
        if "PortfolioPercentWeights" in df_stat["name"].tolist():
            sim_df_raw = df_manager_btfied[
                [
                    "PortfolioPercentWeights",
                    "PortfolioPercentTradingWeightsBestHedgePortfolioActiveNoLibertyCSR",
                    "ActiveRiskBestHedgePortfolioActiveNoLibertyCSR",
                ]
            ]
            sim_df_raw = sim_df_raw.rename(
                columns={
                    "PortfolioPercentWeights": "Weights",
                    "PortfolioPercentTradingWeightsBestHedgePortfolioActiveNoLibertyCSR": "BH wgt",
                    "ActiveRiskBestHedgePortfolioActiveNoLibertyCSR": "BH TE",
                }
            )
        elif "PortfolioWeights" in df_stat["name"].tolist():
            sim_df_raw = df_manager_btfied[
                [
                    "PortfolioWeights",
                    "PortfolioPercentTradingWeightsBestHedgePortfolioActiveNoLibertyCSR",
                    "ActiveRiskBestHedgePortfolioActiveNoLibertyCSR",
                ]
            ]
            sim_df_raw = sim_df_raw.rename(
                columns={
                    "PortfolioWeights": "Weights",
                    "PortfolioPercentTradingWeightsBestHedgePortfolioActiveNoLibertyCSR": "BH wgt",
                    "ActiveRiskBestHedgePortfolioActiveNoLibertyCSR": "BH TE",
                }
            )
        sim_df = sim_df_raw.drop("portfolio")
        sorted_sim_df = sim_df.sort_values("BH TE")

        return sorted_sim_df, cur_port_te

    def manager_weight_simulation_trading(
        self, weights
    ):  # Ipysheet to ask for trading weights from user
        """
        Ipysheet where user provides trading weights
        Args:
            weights(list): Current weights from user
        Returns:
                ipysheet: sheet containing portfolio and weights
        """
        try:
            managers = pd.read_csv(self.manager_structures)
            managers = managers["PortfolioName"].values.tolist()
            new_weights = weights[1:]  # Handling portfolio total weight in the list
            form_weights = [
                f"{weight:.{self.dis_decimal}f}" for weight in new_weights
            ]  # Formatting the weights in Ipysheet
            df = pd.DataFrame(
                {
                    self.managers: managers,
                    self.weights: form_weights,
                    self.tradingwgt: [0] * len(managers),
                }
            )
            trade_weights_sheet = ipysheet.from_dataframe(df)
            for cell in trade_weights_sheet.cells:
                if cell.column_start in [0, 1]:
                    cell.read_only = True

            return trade_weights_sheet
        except Exception as e:
            print(
                f"Check the input field names in {self.manager_structures}. It should be PortfolioName, portfolio and benchmark"
            )

    def manager_weight_simulation_trading_weights_calculator(self, trading_sheet):
        """
        Calculating the final weights after trading weight is provided by user.
        Args:
            trading_sheet(dataframe): Dataframe containing trading weights along with portfolio and current weights
        Returns:
            list: Final weights to run Post-trade report
        """
        trading_sheet.replace("", 0.0, inplace=True)
        current_weight = trading_sheet[self.weights].astype(float).values.tolist()
        trading_weight = trading_sheet[self.tradingwgt].astype(float).values.tolist()
        trading_long = []
        for x in trading_weight:
            if np.exp(x) != 1.0:
                trading_long.append(x)
            else:
                trading_long.append(0)
        trading_long_sum = np.sum(trading_long)
        if trading_long_sum == 0:  # Handling 0 for long positions
            trading_long_sum = 1
        else:
            pass
        trading_long_normalized = np.array([x / trading_long_sum for x in trading_long])
        trading_long_normalized
        trading_short = []
        for x, trade in enumerate(trading_weight):
            if np.exp(trade) == 1.0:
                trading_short.append(current_weight[x])
            else:
                trading_short.append(0)
        trading_short_sum = np.sum(trading_short)
        if trading_short_sum == 0:  # Handling 0 for short positions
            trading_short_sum = 1
        else:
            pass
        trading_short_normalized = np.array(
            [(-1.0 * (x / trading_short_sum)) for x in trading_short]
        )
        trading_short_normalized
        trading_agg = trading_long_normalized + trading_short_normalized
        final_weight = current_weight + trading_agg * trading_long_sum

        return list(final_weight)

    def manager_weights_post(self, manager_weights_post_trade, port_holdings):
        """
        Update the portfolio holdings with trading weights
        Args:
            manager_weights_post_trade(dataframe): Dataframe containing trading weights along with portfolio and current weights
            port_holdings(Dict): Portfolio holdings
        Returns:
            Dict - Updated portfolio holdings containing ids and adjusted weights
        """
        emv_post = self.manager_weight_simulation_trading_weights_calculator(
            manager_weights_post_trade
        )
        emv_post = [float(num) for num in emv_post]
        port_holdings["market_values"] = emv_post

        return port_holdings

    def manager_risk_simulation(self, sorted_sim_bh, curr_te):
        """
        Args:
            sorted_sim_bh (df): Sorted dataframe with latest weights, BH trading weights and BH simulated TE
            curr_te (float): Current portfolio level Contribution to Tracking Error
        Returns:
            DataFrame: data structure with simulated Weights and Tracking error
        """
        if curr_te == 0:
            risk_te_pd = "None"
        else:
            sorted_sim_bh["TE"] = np.array([curr_te] * len(sorted_sim_bh["Weights"]))
            index_list = sorted_sim_bh.index.values.tolist()
            risk_te_pd = pd.DataFrame()
            for x in index_list:
                temp_pd = sorted_sim_bh["Weights"].to_frame()
                weights = [sorted_sim_bh.loc[x, "Weights"]]
                sim_port_holdings = self.get_weights(
                    sorted_sim_bh["Weights"].to_frame().reset_index()
                )
                temp_pd["Trading Weights"] = [sorted_sim_bh.loc[x, "BH wgt"]] + [0] * (
                    len(index_list) - 1
                )
                temp_port_holdings = self.manager_weights_post(
                    temp_pd, sim_port_holdings
                )
                weights.append(temp_port_holdings["market_values"][0])
                te = [sorted_sim_bh.loc[x, "TE"], sorted_sim_bh.loc[x, "BH TE"]]

                if int(sorted_sim_bh["BH wgt"][0]) == 0:
                    pass

                else:
                    # Formulating an equation
                    h, k = (weights[1], te[1])
                    a = (te[0] - k) / (weights[0] - h) ** 2

                    # Simulating the x values to find y
                    num_diff = abs(weights[0] - weights[1]) * 1.75
                    x_values = np.arange(
                        weights[1] - num_diff,
                        weights[1] + num_diff + (1 / (1000000)),
                        num_diff / 10,
                    )
                    y_values = a * (x_values - h) ** 2 + k

                    risk_wgt_change_list = np.linspace(
                        sorted_sim_bh.loc[x, "Weights"], weights[1], 3
                    ).tolist()
                    risk_te_change_list = np.linspace(
                        sorted_sim_bh.loc[x, "BH TE"], curr_te, 3
                    ).tolist()
                    risk_te_pd.loc[x, ["TE", "Weights", "Chng Weights", "Chng TE"]] = (
                        str(y_values.tolist()),
                        str(x_values.tolist()),
                        str(risk_wgt_change_list),
                        str(risk_te_change_list),
                    )
                    sorted_sim_bh = pd.concat(
                        [sorted_sim_bh.iloc[1:], sorted_sim_bh.iloc[:1]]
                    )
        return risk_te_pd

    def bh_charting(self, risk_te_pd, chart_name):
        """
        Args:
            risk_te_pd (df): dataframe with simulated Weights and Tracking error
            chart_name (str): Suffix to the name of the chart that is being generated. Eg. "<Manager name> - <chart_name>"
        Returns:
            plt.charts(png): Saves the charts under cache directory
        """

        if isinstance(risk_te_pd, str):
            plt.clf()
        else:
            curr_dir = os.path.join(os.path.join(os.getcwd(), "cache"))

            if not os.path.exists(curr_dir):
                os.makedirs(curr_dir)  # create directory if doesn't exist

            for manager_names in risk_te_pd.index.tolist():
                temp_weights = ast.literal_eval(
                    risk_te_pd.loc[manager_names, "Weights"]
                )
                temp_te = ast.literal_eval(risk_te_pd.loc[manager_names, "TE"])
                temp_ch_wgt = ast.literal_eval(
                    risk_te_pd.loc[manager_names, "Chng Weights"]
                )
                temp_ch_te = ast.literal_eval(risk_te_pd.loc[manager_names, "Chng TE"])
                temp_ch_wgt_y = [max(temp_ch_te) for x in range(2)]
                temp_ch_te_x = [
                    temp_weights[temp_te.index(min(temp_te))] for x in range(2)
                ]
                low_index = temp_te.index(min(temp_te))

                # Plotting the data points to generate chart
                plt.plot(temp_weights, temp_te)
                lim_x, lim_y = plt.xlim(), plt.ylim()
                plt.plot(
                    [lim_x[0], temp_weights[low_index]],
                    [min(temp_te), min(temp_te)],
                    color="green",
                    linestyle="--",
                )
                plt.plot(
                    [temp_weights[low_index], temp_weights[low_index]],
                    [min(temp_te), lim_y[0]],
                    color="green",
                    linestyle="--",
                )
                plt.plot(
                    [lim_x[0], temp_ch_wgt[0]],
                    [temp_ch_te[-1], temp_ch_te[-1]],
                    color="orange",
                    linestyle="--",
                )
                plt.plot(
                    [temp_ch_wgt[0], temp_ch_wgt[0]],
                    [temp_ch_te[-1], lim_y[0]],
                    color="orange",
                    linestyle="--",
                )
                plt.scatter(
                    temp_weights[low_index],
                    temp_te[low_index],
                    color="green",
                    marker="o",
                )
                plt.scatter(temp_ch_wgt[0], temp_ch_te[-1], color="yellow", marker="v")

                # Setting title and limits
                plt.xlim(lim_x[0], lim_x[1])
                plt.ylim(lim_y[0], lim_y[1])
                plt.title(f"{manager_names} - {chart_name}")
                plt.xlabel("Weight")
                plt.ylabel("Simulated risk")

                arrow_properties = dict(
                    arrowstyle="->", lw=1.25, color="black", mutation_scale=20
                )
                arrow_up = 0.12 * (temp_ch_te[1] - temp_ch_te[0])

                # Annotates
                plt.annotate(
                    "",
                    xy=(temp_ch_wgt[-1], max(temp_ch_te) + arrow_up),
                    xytext=(temp_ch_wgt[0], max(temp_ch_te) + arrow_up),
                    arrowprops=arrow_properties,
                )  # for arrows in the chart
                plt.annotate(
                    "",
                    xy=(temp_weights[temp_te.index(min(temp_te))], temp_ch_te[0]),
                    xytext=(temp_weights[temp_te.index(min(temp_te))], temp_ch_te[-1]),
                    arrowprops=arrow_properties,
                )  # for arrows in the chart
                plt.annotate(
                    textwrap.fill(
                        f"Best Hedge risk level -> {round(min(temp_te),3)}", 15
                    ),
                    (temp_weights[low_index], temp_te[low_index]),
                    fontsize=9,
                    textcoords="offset points",
                    xytext=(128, 0),
                    ha="center",
                    color="blue",
                )
                plt.annotate(
                    textwrap.fill(
                        f"Best-Hedge Trade amount -> {round(abs(temp_ch_wgt[0]- temp_ch_wgt[-1]),3)}%",
                        20,
                    ),
                    (temp_ch_wgt[1], temp_ch_wgt_y[1]),
                    fontsize=9,
                    textcoords="offset points",
                    xytext=(0, 15),
                    ha="center",
                    color="black",
                )
                plt.annotate(
                    textwrap.fill(
                        f"Risk-Hedge  {round(abs(temp_ch_te[0]- temp_ch_te[-1]),3)}%",
                        15,
                    ),
                    (temp_ch_te_x[1], temp_ch_te[1]),
                    fontsize=9,
                    textcoords="offset points",
                    xytext=(5, 0),
                    ha="left",
                    color="black",
                )
                (p1,) = plt.plot(
                    temp_weights, temp_te, label="Trade Simulation Profile"
                )  # Adding the legend
                (p2,) = plt.plot(
                    temp_weights[low_index],
                    temp_te[low_index],
                    color="green",
                    marker="o",
                    label="Best Hedge",
                )
                (p3,) = plt.plot(
                    temp_ch_wgt[0],
                    temp_ch_te[-1],
                    color="orange",
                    marker="v",
                    label="Current Portfolio",
                )

                plt.legend(handles=[p1, p2, p3])

                # Saving the charts
                plt.savefig(
                    os.path.join(curr_dir, f"{manager_names} figure.png"), dpi=300
                )
                plt.clf()

    def process_image(self, img):
        """
        Process the image and return the image in "RGB" mode if it's not.
        Args:
            img (Image): original image
        Return:
            img (Image): processed image
        """
        if img.mode != "RGB":
            temp_img = Image.new("RGB", img.size, (255, 255, 255))
            temp_img.paste(img, mask=img.split()[3])
            img = temp_img
        return img

    def merge_images(self, images, img_dim, grid_dim):
        """
        Merges images together given a specific grid
        Args:
            images (list): list of PIL Image
            img_dim (tuple): dimensions of each image
            grid_dim (tuple): dimensions of grid
        Return:
            final_img (Image): final merged image
        """
        final_img = Image.new(
            "RGB", (img_dim[0] * grid_dim[0], img_dim[1] * grid_dim[1]), (255, 255, 255)
        )
        for i, img in enumerate(images):
            final_img.paste(
                img, ((i % grid_dim[0]) * img_dim[0], (i // grid_dim[0]) * img_dim[1])
            )
        return final_img

    def display_chart_single(self, image_paths):
        """
        Args:
            image_paths (list): list of manager names to display the images
        Returns:

        """
        img_width, img_height = 600, 400  # Dimensions for a single image

        columns = 2 if len(image_paths) > 2 else len(image_paths)
        rows = math.ceil(len(image_paths) / columns)

        # Scale final display size based on the number of images
        final_display_width = img_width * columns
        final_display_height = img_height * rows

        images = [
            self.process_image(
                # Open and resize image to given dimensions
                Image.open(f"./cache/{img} figure.png").resize((img_width, img_height))
            )
            for img in image_paths
        ]

        # Merge images
        final_img = self.merge_images(images, (img_width, img_height), (columns, rows))

        # Create a new figure with scaled size
        plt.figure(
            figsize=(final_display_width / 80, final_display_height / 80)
        )  # Divide by 80 or suitable number to convert pixels to inches

        plt.imshow(final_img)
        plt.axis("off")  # Do not show axis
        plt.show()

    def display_chart(self, sorted_sim_bh):
        """
        Args:
            sorted_sim_bh (df): Sorted dataframe with latest weights, BH trading weights and BH simulated TE
        Returns:
            ipywidgets: Displays the UI to filter the charts that user wants
        """
        # Beginning of your function, just before the process starts

        if sorted_sim_bh["Weights"].tolist() == [0] * len(
            sorted_sim_bh["Weights"].tolist()
        ):
            print("You have entered no weights to generate risk analytics and charts.")
        else:
            index_list = sorted_sim_bh.index.tolist()
            for x in index_list:
                if int(sorted_sim_bh.loc[x, "BH wgt"]) == 0:
                    sorted_sim_bh = sorted_sim_bh.drop(x)
            index_list = sorted_sim_bh.index.tolist()

            main_options = {
                "None": None,
                "Bottom 2 Risk-hedge": None,
                "Top 2 Risk-hedge": None,
                "Manager-wise": sorted(index_list),
            }
            main_dropdown = widgets.Dropdown(
                options=main_options.keys(), description="Options", value="None"
            )
            multi_select = widgets.SelectMultiple(description="Managers:")

            # Button to generate output
            generate_button = widgets.Button(description="Display Chart")
            output = widgets.Output()

            # Update the sub_dropdown options based on main_dropdown selection
            def update_sub_options(*args):
                sub_options = main_options[main_dropdown.value]
                if sub_options is None:
                    multi_select.layout.display = "none"
                    generate_button.layout.display = "none"
                    generate_output()
                else:
                    multi_select.layout.display = "flex"
                    generate_button.layout.display = "flex"
                    multi_select.options = sub_options
                    output.clear_output()

            value = main_dropdown.observe(update_sub_options)

            # Functionality of output generation
            def generate_output(*args):
                with output:
                    output.clear_output()
                    if main_dropdown.value == "Bottom 2 Risk-hedge":
                        img1 = Image.open(f"./cache/{index_list[-1]} figure.png")
                        img2 = Image.open(f"./cache/{index_list[-2]} figure.png")
                        combined_image = Image.new(
                            "RGB",
                            (img1.width + img2.width, max(img1.height, img2.height)),
                        )
                        combined_image.paste(img1, (0, 0))
                        combined_image.paste(img2, (img1.width, 0))
                        display(combined_image)
                    elif main_dropdown.value == "Top 2 Risk-hedge":
                        img1 = Image.open(f"./cache/{index_list[0]} figure.png")
                        img2 = Image.open(f"./cache/{index_list[1]} figure.png")
                        combined_image = Image.new(
                            "RGB",
                            (img1.width + img2.width, max(img1.height, img2.height)),
                        )
                        combined_image.paste(img1, (0, 0))
                        combined_image.paste(img2, (img1.width, 0))
                        display(combined_image)
                    elif main_dropdown.value == "None":
                        print("")
                    elif main_dropdown.value is not None:
                        n = 1
                        if multi_select.value is list():
                            list_ms = multi_select.value
                        else:
                            list_ms = list(multi_select.value)
                        self.display_chart_single(list_ms)

            generate_button.on_click(generate_output)

            # Display the widgets
            display(
                widgets.HBox([main_dropdown, multi_select, generate_button]), output
            )

            # Ensure that sub_dropdown options are updated when notebook is loaded
            update_sub_options()


class OpenRiskEngine:
    def __init__(self, username, apikey):
        """
        self.username: Username for client's authentication - FDS Username
        self.apikey: apikey associated with FDS Username
        """
        self.username = username
        self.apikey = apikey

    def composite_asset_path_reader(self, ca_port_map):
        try:
            data = []
            for index, row in ca_port_map.iterrows():
                if pd.isnull(row["filepath"]):
                    filepath = row["filename"]
                else:
                    filepath = os.path.join(str(row["filepath"]), str(row["filename"]))

                portfolio = row["portfolio"]

                # Appending a tuple of portfolio and filepath to data list
                data.append((portfolio, filepath))

            # Converting list data to dataframe
            portfolio_filepaths = pd.DataFrame(data, columns=["portfolio", "filepaths"])
            return portfolio_filepaths
        except Exception as e:
            print(
                f"Check the input field names in mapping file. It should be portfolio, filepath and filename"
            )
            raise e

    def composite_asset_builder_expansion(
        self, flat_file, ca_port_map_raw=None, ca_port_map_file=None, ca_expand_lvl=None
    ):
        """
        Parses into various csv holdings files/databases to add holding composite to existing managers structure flat file.
        Args:
            flat_file(df): Flat file from client consisting of composite asset information
            ca_port_map_raw(df): Dataframe consisting of portfolio filename and file path for composite level expansion (optional)
            ca_port_map_file(dict): Contains portfolio name and portfolio dataframe as key and value respectively (optional)
            ca_expand_lvl(int): Composite asset expand level. Mainly used when using composite level expansion. (optional)
        Returns:
                DataFrame: flat file consisting of Composite Asset information
        """
        ca_port_map = self.composite_asset_path_reader(
            ca_port_map_raw
        )  # Adding composite asset path reader to expansion function

        manager_holdings = flat_file["ids"].tolist()
        if ca_port_map is not None:
            ca_port_map.set_index("portfolio", inplace=True)
            ca_port_map_values = ca_port_map.index.values.tolist()
        if ca_port_map_file is not None:
            ca_port_map_file_keys = list(ca_port_map_file.keys())

        if ca_expand_lvl != None:
            temp_holdings_list = []
            for lvl in range(ca_expand_lvl + 1):
                for x in manager_holdings:
                    if (
                        ca_port_map is not None and x in ca_port_map_values
                    ):  # Using the csv files
                        file = ca_port_map.loc[x, "filepaths"]
                        df_temp = pd.read_csv(file)
                        df_temp["compositeAssetId"] = x
                        flat_file = pd.concat([flat_file, df_temp], ignore_index=True)
                        temp_holdings_list.append(x)
                    elif (
                        ca_port_map_file != None and x in ca_port_map_file_keys
                    ):  # Using the mapped dictonary with portfolio data
                        df_temp = ca_port_map_file[x]
                        df_temp["compositeAssetId"] = x
                        flat_file = pd.concat([flat_file, df_temp], ignore_index=True)
                        temp_holdings_list.append(x)
                manager_holdings = list(set(manager_holdings) - set(temp_holdings_list))
                new_holdings_list = flat_file["ids"].tolist()
                new_holdings_list = list(
                    set(new_holdings_list) - set(temp_holdings_list)
                )
                manager_holdings = list(
                    set(manager_holdings).union(set(new_holdings_list))
                )

        flat_file.dropna(inplace=True)
        return flat_file

    def composite_asset_builder_esdk(self, flat_file):
        """
        Converts the flat file into ESDK compatible composite asset format.
        Args:
            flat_file(Dataframe): Flat file consisting of Composite Asset information
        Returns:
                sdk.composite_asset_definitions
        """
        flat_file.set_index("compositeAssetId", inplace=True)

        asset_definitions = {
            name: IDsAndMarketValues(
                ids=flat_file["ids"].astype(str).tolist(),
                market_values=flat_file["marketValues"].astype(float).tolist(),
            )
            for name, flat_file in flat_file.groupby("compositeAssetId")
        }

        composite_assets = CompositeAssetDefinitions(**asset_definitions)

        return composite_assets

    def stat_remap(self, output_dict, stat_remapping_file):
        """
        Stat remapping file renames the general API stat back to the remapped one in the csv file
        Args:
            output_dict (Dict) : Dictionary to rename
            stat_remapping_file (str): csv path where file contains information on mapping statistics
        Returns:
            Dict: Dictionary with remapped keys
        """
        try:
            stat_remapping = pd.read_csv(stat_remapping_file)
            if not {"statname", "name"}.issubset(stat_remapping.columns):
                print(
                    "Columns 'statname' and 'name' not found in CSV file. Returning original dictionary."
                )
                return output_dict
            remap_dict = stat_remapping.set_index("statname")["name"].to_dict()
            for key in output_dict:
                if output_dict[key] is not None:
                    output_dict[key].columns = (
                        output_dict[key].columns.to_series().replace(remap_dict)
                    )
            return output_dict
        except FileNotFoundError:
            print(
                "File not found: "
                + stat_remapping_file
                + ". Returning original dictionary."
            )
            return output_dict
        except Exception as e:
            print(f"An error occurred: {e}. Returning original dictionary.")

            return output_dict

    def raw_stat_sdk(self, df_stat):
        """
        Converts Stats to ESDK compatiable input-stats
        Args:
            df_stat(df): Dataframe containing the statistics information
        Returns:
            Stats: A Statistics object that can be used by SDK
        """
        try:
            stats = []
            for i in df_stat["name"].keys():
                name = df_stat["name"][i]
                level = StatCalculationLevel(df_stat["level"][i])
                settings_dict = {}
                for k in df_stat:
                    if k not in ["name", "level"]:
                        try:
                            if not np.isnan(df_stat[k][i]):
                                settings_dict[k] = df_stat[k][i]
                        except:
                            settings_dict[k] = df_stat[k][i]
                # Only create 'settings' object if settings_dict is not empty
                if settings_dict:
                    settings = StatCalculationSettings(**settings_dict)
                    stats.append(Stat(name=name, level=level, settings=settings))
                else:
                    stats.append(Stat(name=name, level=level))

            return stats
        except Exception as e:
            print(
                "Check field names in stats csv. 'name' and 'level' are compulsory inputs."
            )

    def sdk_calculate_from_holdings_request(
        self,
        date,
        port_holdings,
        currency,
        calendar,
        risk_model,
        df_stat,
        composite_assets,
    ):
        """
        SDK Request to Open:Risk API
        Args:
            date (str): Portfolio date
            port_holdings: Portfolio holdings - contains ids and marketValues
            bench_holdings: Benchmark holdings - contains ids and marketValues
            currency (str): report currency
            calendar (str): report calendar code
            risk_model (str): RiskModel ID
            df_stat: stats object, should contain stat_name and stat_level
            composite_assets: composite asset definition
        Returns:
            dict: Returns the response from calculate/from-Holdings route
        """
        try:
            # Setting-up Stats for ESDK
            raw_stats = self.raw_stat_sdk(df_stat)

            # Setting-up request body
            configuration = fds.sdk.OpenRisk.Configuration(
                username=self.username, password=self.apikey
            )
            with fds.sdk.OpenRisk.ApiClient(configuration) as api_client:
                api_instance = operations_api.OperationsApi(api_client)
                version = "v1"
                calculate_from_holdings_request_body = CalculateFromHoldingsRequestBody(
                    data=CalculateFromHoldingsRequestData(
                        date=Date(date),
                        holdings=Holdings(
                            portfolio=Holding(**port_holdings),
                        ),
                        currency=CurrencyISOCode(currency),
                        calendar=CalendarCode(calendar),
                        risk_model=RiskModelCode(risk_model),
                        stats=Stats(raw_stats),
                        composite_assets=composite_assets,
                    )
                )

            api_response = api_instance.calculate_from_holdings(
                version, calculate_from_holdings_request_body
            )
            return api_response
        except Exception as e:
            print("Check the inputs procided in stats and risk variables")
            raise e

    def calculate_response_json_to_df_single_date(self, json_openrisk_output):
        """
        Converts Open Risk API JSON output into DataFrame for a specific date for each different levels
        """
        df_portfolio = None
        df_security = None
        df_factor = None
        df_factorgroup = None
        df_factorfactor = None
        df_factorsecurity = None
        df_securitysecurity = None
        for y, statlevel in enumerate(json_openrisk_output["data"]["value"]):
            if statlevel["level"] == "Security":
                df_dict = {
                    "symbol": json_openrisk_output["meta"]["labels"]["security"]["ids"],
                    "date": [
                        json_openrisk_output["meta"]["resolved_dates"]["universe_date"]
                    ]
                    * len(json_openrisk_output["meta"]["labels"]["security"]["ids"]),
                    statlevel["stat"]: statlevel["value"],
                }
                if df_security is None:
                    df_security = pd.DataFrame(df_dict)
                else:
                    temp_col = statlevel["value"]
                    df_security[statlevel["stat"]] = temp_col
            elif statlevel["level"] == "Portfolio":
                df_dict = {
                    "portfolio": ["portfolio"],
                    "date": [
                        json_openrisk_output["meta"]["resolved_dates"]["universe_date"]
                    ],
                    statlevel["stat"]: statlevel["value"],
                }
                if df_portfolio is None:
                    df_portfolio = pd.DataFrame(df_dict)
                else:
                    temp_col = statlevel["value"]
                    df_portfolio[statlevel["stat"]] = temp_col
            elif statlevel["level"] == "Factor":
                factorIds = json_openrisk_output["meta"]["labels"]["factor"]["ids"]
                df_dict = {
                    "factor": factorIds,
                    "date": [
                        json_openrisk_output["meta"]["resolved_dates"]["universe_date"]
                    ]
                    * len(factorIds),
                    statlevel["stat"]: statlevel["value"],
                }
                if df_factor is None:
                    df_factor = pd.DataFrame(df_dict)
                else:
                    temp_col = statlevel["value"]
                    df_factor[statlevel["stat"]] = temp_col
            elif statlevel["level"] == "FactorGroup":
                factorGroups = []
                for x in json_openrisk_output["meta"]["labels"]["factorGroup"][
                    "groups"
                ]:
                    factorGroups.append(x["name"])
                df_dict = {
                    "factorGroups": factorGroups,
                    "date": [
                        json_openrisk_output["meta"]["resolved_dates"]["universe_date"]
                    ]
                    * len(factorGroups),
                    statlevel["stat"]: statlevel["value"],
                }
                if df_factorgroup is None:
                    df_factorgroup = pd.DataFrame(df_dict)
                else:
                    temp_col = statlevel["value"]
                    df_factorgroup[json_openrisk_output["data"][y]["stat"]] = temp_col
            elif statlevel["level"] == "FactorFactor":
                factorIds = json_openrisk_output["meta"]["labels"]["factor"]["ids"]
                factorFactorValues = statlevel["value"]
                if isinstance(factorFactorValues, dict):
                    if df_factorfactor is None:
                        df_factorfactor = pd.DataFrame(
                            csr_matrix(
                                (
                                    factorFactorValues["value"],
                                    factorFactorValues["indices"],
                                    factorFactorValues["indexPointer"],
                                ),
                                shape=(
                                    factorFactorValues["rows"],
                                    factorFactorValues["columns"],
                                ),
                            ).toarray()
                        )
                        df_factorfactor.insert(0, "", factorIds)
                        df_factorfactor.insert(
                            0,
                            "date",
                            [
                                json_openrisk_output["meta"]["resolved_dates"][
                                    "universe_date"
                                ]
                            ]
                            * len(factorIds),
                        )
                        df_factorfactor.columns = [
                            "date",
                            statlevel["stat"],
                        ] + factorIds
                    else:
                        temp_df = pd.DataFrame(
                            csr_matrix(
                                (
                                    factorFactorValues["value"],
                                    factorFactorValues["indices"],
                                    factorFactorValues["indexPointer"],
                                ),
                                shape=(
                                    factorFactorValues["rows"],
                                    factorFactorValues["columns"],
                                ),
                            ).toarray()
                        )
                        temp_df.insert(0, "", factorIds)
                        temp_df.columns = [statlevel["stat"]] + factorIds
                        df_factorfactor = pd.concat([df_factorfactor, temp_df], axis=1)
                if isinstance(factorFactorValues, list):
                    if df_factorfactor is None:
                        df_factorfactor = pd.DataFrame(factorFactorValues)
                        df_factorfactor.insert(0, "", factorIds)
                        df_factorfactor.insert(
                            0,
                            "date",
                            [
                                json_openrisk_output["meta"]["resolved_dates"][
                                    "universe_date"
                                ]
                            ]
                            * len(factorIds),
                        )
                        df_factorfactor.columns = [
                            "date",
                            statlevel["stat"],
                        ] + factorIds
                    else:
                        temp_df = pd.DataFrame(factorFactorValues)
                        temp_df.insert(0, "", factorIds)
                        temp_df.columns = [statlevel["stat"]] + factorIds
                        df_factorfactor = pd.concat([df_factorfactor, temp_df], axis=1)
            elif statlevel["level"] == "FactorSecurity":
                symbols = json_openrisk_output["meta"]["labels"]["security"]["ids"]
                factorIds = json_openrisk_output["meta"]["labels"]["factor"]["ids"]
                factorSecurityValues = statlevel["value"]
                if isinstance(factorSecurityValues, dict):
                    if df_factorsecurity is None:
                        df_factorsecurity = pd.DataFrame(
                            csr_matrix(
                                (
                                    factorSecurityValues["value"],
                                    factorSecurityValues["indices"],
                                    factorSecurityValues["indexPointer"],
                                ),
                                shape=(
                                    factorSecurityValues["rows"],
                                    factorSecurityValues["columns"],
                                ),
                            ).toarray()
                        ).transpose()
                        df_factorsecurity.insert(0, "symbol", symbols)
                        df_factorsecurity.insert(
                            0,
                            "date",
                            [
                                json_openrisk_output["meta"]["resolved_dates"][
                                    "universe_date"
                                ]
                            ]
                            * len(symbols),
                        )
                        df_factorsecurity.columns = ["date", "symbol"] + factorIds
                    else:
                        temp_df = pd.DataFrame(
                            csr_matrix(
                                (
                                    factorSecurityValues["value"],
                                    factorSecurityValues["indices"],
                                    factorSecurityValues["indexPointer"],
                                ),
                                shape=(
                                    factorSecurityValues["rows"],
                                    factorSecurityValues["columns"],
                                ),
                            ).toarray()
                        ).transpose()
                        temp_df.insert(0, statlevel["stat"], symbols)
                        temp_df.columns = [statlevel["stat"]] + factorIds
                        df_factorsecurity = pd.concat(
                            [df_factorsecurity, temp_df], axis=1
                        )
                if isinstance(factorSecurityValues, list):
                    if df_factorsecurity is None:
                        df_factorsecurity = pd.DataFrame(factorSecurityValues)
                        df_factorsecurity = df_factorsecurity.transpose()
                        df_factorsecurity.insert(0, "symbol", symbols)
                        df_factorsecurity.insert(
                            0,
                            "date",
                            [
                                json_openrisk_output["meta"]["resolved_dates"][
                                    "universe_date"
                                ]
                            ]
                            * len(symbols),
                        )
                        df_factorsecurity.columns = ["date", "symbol"] + factorIds
                    else:
                        temp_df = pd.DataFrame(factorSecurityValues)
                        temp_df = temp_df.transpose()
                        temp_df.insert(0, statlevel["stat"], symbols)
                        temp_df.columns = [statlevel["stat"]] + factorIds
                        df_factorsecurity = pd.concat(
                            [df_factorsecurity, temp_df], axis=1
                        )
            elif statlevel["level"] == "SecuritySecurity":
                securitysecurityvalue = statlevel["value"]
                symbols = json_openrisk_output["meta"]["labels"]["security"]["ids"]
                if isinstance(securitysecurityvalue, dict):
                    if df_securitysecurity is None:
                        df_securitysecurity = pd.DataFrame(
                            csr_matrix(
                                (
                                    securitysecurityvalue["value"],
                                    securitysecurityvalue["indices"],
                                    securitysecurityvalue["indexPointer"],
                                ),
                                shape=(
                                    securitysecurityvalue["rows"],
                                    securitysecurityvalue["columns"],
                                ),
                            ).toarray()
                        )
                        df_securitysecurity.insert(0, "", symbols)
                        df_securitysecurity.insert(
                            0,
                            "date",
                            [
                                json_openrisk_output["meta"]["resolved_dates"][
                                    "universe_date"
                                ]
                            ]
                            * len(symbols),
                        )
                        df_securitysecurity.columns = [
                            "date",
                            statlevel["stat"],
                        ] + symbols
                    else:
                        temp_df = pd.DataFrame(
                            csr_matrix(
                                (
                                    securitysecurityvalue["value"],
                                    securitysecurityvalue["indices"],
                                    securitysecurityvalue["indexPointer"],
                                ),
                                shape=(
                                    securitysecurityvalue["rows"],
                                    securitysecurityvalue["columns"],
                                ),
                            ).toarray()
                        )
                        temp_df.insert(0, "", symbols)
                        temp_df.insert(
                            0,
                            "date",
                            [
                                json_openrisk_output["meta"]["resolved_dates"][
                                    "universe_date"
                                ]
                            ]
                            * len(symbols),
                        )
                        temp_df.columns = ["date", statlevel["stat"]] + symbols
                        df_securitysecurity = pd.concat(
                            [df_securitysecurity, temp_df], axis=1
                        )
                if isinstance(securitysecurityvalue, list):
                    if df_securitysecurity is None:
                        df_securitysecurity = pd.DataFrame(securitysecurityvalue)
                        df_securitysecurity.insert(0, "", symbols)
                        df_securitysecurity.insert(
                            0,
                            "date",
                            [
                                json_openrisk_output["meta"]["resolved_dates"][
                                    "universe_date"
                                ]
                            ]
                            * len(symbols),
                        )
                        df_securitysecurity.columns = [
                            "date",
                            statlevel["stat"],
                        ] + symbols
                    else:
                        temp_df = pd.DataFrame(securitysecurityvalue)
                        temp_df.insert(0, "", symbols)
                        temp_df.columns = [statlevel["stat"]] + symbols
                        df_securitysecurity = pd.concat(
                            [df_securitysecurity, temp_df], axis=1
                        )
                # replace None values in dataframe with 0.0
        if df_portfolio is not None:
            df_portfolio = df_portfolio.fillna(value=0.0)
        if df_security is not None:
            df_security = df_security.fillna(value=0.0)
        if df_factor is not None:
            df_factor = df_factor.fillna(value=0.0)
        if df_factorgroup is not None:
            df_factorgroup = df_factorgroup.fillna(value=0.0)
        if df_factorfactor is not None:
            df_factorfactor = df_factorfactor.fillna(value=0.0)
        if df_factorsecurity is not None:
            df_factorsecurity = df_factorsecurity.fillna(value=0.0)
        if df_securitysecurity is not None:
            df_securitysecurity = df_securitysecurity.fillna(value=0.0)

        return {
            "portfolio": df_portfolio,
            "security": df_security,
            "factor": df_factor,
            "factorGroup": df_factorgroup,
            "factorfactor": df_factorfactor,
            "factorsecurity": df_factorsecurity,
            "securitysecurity": df_securitysecurity,
        }
